@@ -23,6 +23,7 @@
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+#include "exported-buffer-shm-private.h"
 #include "view-backend-exportable-private.h"
 #include "ws.h"
 #include <cassert>
@@ -105,6 +106,14 @@ public:
         client->export_dmabuf_resource(data, &dmabuf_resource);
     }
 
+    void exportBuffer(struct wl_resource* bufferResource, struct wl_shm_buffer* shmBuffer) override
+    {
+        auto* buffer = new struct wpe_fdo_shm_exported_buffer;
+        buffer->resource = bufferResource;
+        buffer->shm_buffer = shmBuffer;
+        client->export_shm_buffer(data, buffer);
+    }
+
     void releaseBuffer(struct wl_resource* buffer)
     {
         BufferResource* matchingResource = nullptr;
@@ -126,6 +135,13 @@ public:
         delete matchingResource;
     }
 
+    void releaseBuffer(struct wpe_fdo_shm_exported_buffer* buffer)
+    {
+        if (buffer->resource)
+            viewBackend->releaseBuffer(buffer->resource);
+        delete buffer;
+    }
+
     const struct wpe_view_backend_exportable_fdo_client* client;
 
     struct wl_list bufferResources;
@@ -188,4 +204,11 @@ wpe_view_backend_exportable_fdo_dispatch_release_buffer(struct wpe_view_backend_
     static_cast<ClientBundleBuffer*>(exportable->clientBundle)->releaseBuffer(buffer);
 }
 
+__attribute__((visibility("default")))
+void
+wpe_view_backend_exportable_fdo_dispatch_release_shm_exported_buffer(struct wpe_view_backend_exportable_fdo* exportable, struct wpe_fdo_shm_exported_buffer* buffer)
+{
+    static_cast<ClientBundleBuffer*>(exportable->clientBundle)->releaseBuffer(buffer);
+}
+
 }
