@@ -27,9 +27,9 @@
 
 #include <EGL/egl.h>
 #include <EGL/eglext.h>
-#include <EGL/eglmesaext.h>
 #include "linux-dmabuf/linux-dmabuf.h"
 #include "bridge/wpe-bridge-server-protocol.h"
+#include "video-plane-display-dmabuf/wpe-video-plane-display-dmabuf-server-protocol.h"
 #include <cassert>
 #include <stdio.h>
 #include <string.h>
@@ -39,6 +39,7 @@
 #ifndef EGL_WL_bind_wayland_display
 #define EGL_WAYLAND_BUFFER_WL 0x31D5
 typedef EGLBoolean (EGLAPIENTRYP PFNEGLBINDWAYLANDDISPLAYWL) (EGLDisplay dpy, struct wl_display *display);
+typedef EGLBoolean (EGLAPIENTRYP PFNEGLQUERYWAYLANDBUFFERWL) (EGLDisplay dpy, struct wl_resource *buffer, EGLint attribute, EGLint *value);
 #endif
 
 #ifndef EGL_EXT_image_dma_buf_import_modifiers
@@ -80,6 +81,10 @@ static PFNEGLDESTROYIMAGEKHRPROC s_eglDestroyImageKHR;
 static PFNEGLQUERYDMABUFFORMATSEXTPROC s_eglQueryDmaBufFormatsEXT;
 static PFNEGLQUERYDMABUFMODIFIERSEXTPROC s_eglQueryDmaBufModifiersEXT;
 
+struct wpe_video_plane_display_dmabuf_export {
+    struct wl_resource* updateResource;
+};
+
 namespace WS {
 
 struct ServerSource {
@@ -112,7 +117,7 @@ GSourceFuncs ServerSource::s_sourceFuncs = {
 
         if (source.pfd.revents & G_IO_IN) {
             struct wl_event_loop* eventLoop = wl_display_get_event_loop(source.display);
-            wl_event_loop_dispatch(eventLoop, -1);
+            wl_event_loop_dispatch(eventLoop, 0);
             wl_display_flush_clients(source.display);
         }
 
@@ -135,6 +140,7 @@ struct Surface {
 
     struct wl_resource* bufferResource { nullptr };
     const struct linux_dmabuf_buffer* dmabufBuffer { nullptr };
+    struct wl_shm_buffer* shmBuffer { nullptr };
 };
 
 static const struct wl_surface_interface s_surfaceInterface = {
@@ -146,6 +152,7 @@ static const struct wl_surface_interface s_surfaceInterface = {
         auto& surface = *static_cast<Surface*>(wl_resource_get_user_data(surfaceResource));
 
         surface.dmabufBuffer = Instance::singleton().getDmaBufBuffer(bufferResource);
+        surface.shmBuffer = wl_shm_buffer_get(bufferResource);
 
         if (surface.bufferResource)
             wl_buffer_send_release(surface.bufferResource);
@@ -185,6 +192,8 @@ static const struct wl_surface_interface s_surfaceInterface = {
 
         if (surface.dmabufBuffer)
             surface.exportableClient->exportLinuxDmabuf(surface.dmabufBuffer);
+        else if (surface.shmBuffer)
+            surface.exportableClient->exportShmBuffer(bufferResource, surface.shmBuffer);
         else
             surface.exportableClient->exportBufferResource(bufferResource);
     },
@@ -192,7 +201,7 @@ static const struct wl_surface_interface s_surfaceInterface = {
     [](struct wl_client*, struct wl_resource*, int32_t) { },
     // set_buffer_scale
     [](struct wl_client*, struct wl_resource*, int32_t) { },
-#ifdef WAYLAND_1_10_OR_GREATER
+#if (WAYLAND_VERSION_MAJOR > 1) || (WAYLAND_VERSION_MAJOR == 1 && WAYLAND_VERSION_MINOR >= 10)
     // damage_buffer
     [](struct wl_client*, struct wl_resource*, int32_t, int32_t, int32_t, int32_t) { },
 #endif
@@ -238,6 +247,51 @@ static const struct wpe_bridge_interface s_wpeBridgeInterface = {
     },
 };
 
+struct DmaBufUpdate {
+    uint32_t id { 0 };
+    struct wl_client* client;
+};
+
+static const struct wpe_video_plane_display_dmabuf_update_interface s_videoPlaneDisplayUpdateInterface = {
+    // destroy
+    [](struct wl_client*, struct wl_resource* resource)
+    {
+        wl_resource_destroy(resource);
+    },
+};
+
+static const struct wpe_video_plane_display_dmabuf_interface s_wpeDmaBufInterface = {
+    // create_update
+    [](struct wl_client* client, struct wl_resource* resource, uint32_t id, uint32_t video_id, int32_t fd, int32_t x, int32_t y, int32_t width, int32_t height, uint32_t stride)
+    {
+        struct wl_resource* updateResource = wl_resource_create(client, &wpe_video_plane_display_dmabuf_update_interface,
+            wl_resource_get_version(resource), id);
+        if (!updateResource) {
+            wl_resource_post_no_memory(resource);
+            return;
+        }
+
+        auto* update = new DmaBufUpdate;
+        update->id = id;
+        update->client = client;
+        wl_resource_set_implementation(updateResource, &s_videoPlaneDisplayUpdateInterface, update,
+            [](struct wl_resource* resource)
+            {
+                auto* update = static_cast<DmaBufUpdate*>(wl_resource_get_user_data(resource));
+                delete update;
+            });
+
+        auto* dmabuf_export = new struct wpe_video_plane_display_dmabuf_export;
+        dmabuf_export->updateResource = updateResource;
+        Instance::singleton().handleVideoPlaneDisplayDmaBuf(dmabuf_export, video_id, fd, x, y, width, height, stride);
+    },
+    // end_of_stream
+    [](struct wl_client* client, struct wl_resource* resource, uint32_t video_id)
+    {
+        Instance::singleton().handleVideoPlaneDisplayDmaBufEndOfStream(video_id);
+    },
+};
+
 Instance& Instance::singleton()
 {
     static Instance* s_singleton;
@@ -315,6 +369,9 @@ Instance::~Instance()
         wl_global_destroy(m_linuxDmabuf);
     }
 
+    if (m_videoPlaneDisplayDmaBuf.object)
+        wl_global_destroy(m_videoPlaneDisplayDmaBuf.object);
+
     if (m_display)
         wl_display_destroy(m_display);
 }
@@ -340,10 +397,14 @@ bool Instance::initialize(EGLDisplay eglDisplay)
         return true;
 
     if (m_eglDisplay != EGL_NO_DISPLAY) {
-        fprintf(stderr, "WPE fdo doesn't support multiple EGL displays\n");
+        g_warning("Multiple EGL displays are not supported.\n");
         return false;
     }
 
+    // wl_display_init_shm() returns `0` on success.
+    if (wl_display_init_shm(m_display) != 0)
+        return false;
+
     const char* extensions = eglQueryString(eglDisplay, EGL_EXTENSIONS);
     if (isEGLExtensionSupported(extensions, "EGL_WL_bind_wayland_display")) {
         s_eglBindWaylandDisplayWL = reinterpret_cast<PFNEGLBINDWAYLANDDISPLAYWL>(eglGetProcAddress("eglBindWaylandDisplayWL"));
@@ -351,8 +412,6 @@ bool Instance::initialize(EGLDisplay eglDisplay)
         s_eglQueryWaylandBufferWL = reinterpret_cast<PFNEGLQUERYWAYLANDBUFFERWL>(eglGetProcAddress("eglQueryWaylandBufferWL"));
         assert(s_eglQueryWaylandBufferWL);
     }
-    if (!s_eglBindWaylandDisplayWL || !s_eglQueryWaylandBufferWL)
-        return false;
 
     if (isEGLExtensionSupported(extensions, "EGL_KHR_image_base")) {
         s_eglCreateImageKHR = reinterpret_cast<PFNEGLCREATEIMAGEKHRPROC>(eglGetProcAddress("eglCreateImageKHR"));
@@ -360,11 +419,13 @@ bool Instance::initialize(EGLDisplay eglDisplay)
         s_eglDestroyImageKHR = reinterpret_cast<PFNEGLDESTROYIMAGEKHRPROC>(eglGetProcAddress("eglDestroyImageKHR"));
         assert(s_eglDestroyImageKHR);
     }
-    if (!s_eglCreateImageKHR || !s_eglDestroyImageKHR)
-        return false;
 
-    if (!s_eglBindWaylandDisplayWL(eglDisplay, m_display))
-        return false;
+    if (s_eglBindWaylandDisplayWL && s_eglQueryWaylandBufferWL) {
+        if (!s_eglCreateImageKHR || !s_eglDestroyImageKHR)
+            return false;
+        if (!s_eglBindWaylandDisplayWL(eglDisplay, m_display))
+            return false;
+    }
 
     m_eglDisplay = eglDisplay;
 
@@ -506,7 +567,10 @@ void Instance::importDmaBufBuffer(struct linux_dmabuf_buffer* buffer)
 
 const struct linux_dmabuf_buffer* Instance::getDmaBufBuffer(struct wl_resource* bufferResource) const
 {
-    if (!m_linuxDmabuf)
+    if (!m_linuxDmabuf || !bufferResource)
+        return nullptr;
+
+    if (!linux_dmabuf_buffer_implements_resource(bufferResource))
         return nullptr;
 
     struct linux_dmabuf_buffer* buffer;
@@ -547,11 +611,55 @@ void Instance::foreachDmaBufModifier(std::function<void (int format, uint64_t mo
     }
 }
 
+void Instance::initializeVideoPlaneDisplayDmaBuf(VideoPlaneDisplayDmaBufCallback updateCallback, VideoPlaneDisplayDmaBufEndOfStreamCallback endOfStreamCallback)
+{
+    if (m_videoPlaneDisplayDmaBuf.object)
+        return;
+
+    m_videoPlaneDisplayDmaBuf.object = wl_global_create(m_display, &wpe_video_plane_display_dmabuf_interface, 1, this,
+        [](struct wl_client* client, void*, uint32_t version, uint32_t id)
+        {
+            struct wl_resource* resource = wl_resource_create(client, &wpe_video_plane_display_dmabuf_interface, version, id);
+            if (!resource) {
+                wl_client_post_no_memory(client);
+                return;
+            }
+
+            wl_resource_set_implementation(resource, &s_wpeDmaBufInterface, nullptr, nullptr);
+        });
+    m_videoPlaneDisplayDmaBuf.updateCallback = updateCallback;
+    m_videoPlaneDisplayDmaBuf.endOfStreamCallback = endOfStreamCallback;
+}
+
+void Instance::handleVideoPlaneDisplayDmaBuf(struct wpe_video_plane_display_dmabuf_export* dmabuf_export, uint32_t id, int fd, int32_t x, int32_t y, int32_t width, int32_t height, uint32_t stride)
+{
+    if (!m_videoPlaneDisplayDmaBuf.updateCallback) {
+        if (fd >= 0)
+            close(fd);
+        return;
+    }
+
+    m_videoPlaneDisplayDmaBuf.updateCallback(dmabuf_export, id, fd, x, y, width, height, stride);
+}
+
+void Instance::handleVideoPlaneDisplayDmaBufEndOfStream(uint32_t id)
+{
+    if (!m_videoPlaneDisplayDmaBuf.endOfStreamCallback)
+        return;
+
+    m_videoPlaneDisplayDmaBuf.endOfStreamCallback(id);
+}
+
+void Instance::releaseVideoPlaneDisplayDmaBufExport(struct wpe_video_plane_display_dmabuf_export* dmabuf_export)
+{
+    wpe_video_plane_display_dmabuf_update_send_release(dmabuf_export->updateResource);
+}
+
 struct wl_client* Instance::registerViewBackend(uint32_t surfaceId, ExportableClient& exportableClient)
 {
     auto it = m_viewBackendMap.find(surfaceId);
     if (it == m_viewBackendMap.end())
-        std::abort();
+        g_error("Instance::registerViewBackend(): " "Cannot find surface %" PRIu32 " in view backend map.", surfaceId);
 
     it->second->exportableClient = &exportableClient;
     return it->second->client;
